//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  func.cpp
//! \brief Function wrapper classes CTypedFunc and CFunc.

#include "defs.hpp"

#include <algorithm>

#include "coord.hpp"
#include "func.hpp"


//**************************************************************************************************
//*  Class CTypedFunc
//**************************************************************************************************

//! Returns the typed function that takes exactly the given input types.

const CTypedFunc* CFunc::find_exact_tyfu(string intypes) const
{
    for (auto p = tyfuList.begin(); p != tyfuList.end(); ++p) {
        if ((*p)->intypes == intypes)
            return *p;
    }
    return NULL;
}

//! Returns the typed function that can deal best with the given input.

const CTypedFunc* CFunc::find_tyfu(string intypes, bool want_error) const
{
    const CTypedFunc* tf;
    if (tf = find_exact_tyfu(intypes))
        return tf;
    // cast all i to d:
    std::replace(intypes.begin(), intypes.end(), 'i', 'd');
    if (tf = find_exact_tyfu(intypes))
        return tf;
    if (intypes.find('e') != string::npos) {
        // cast all d to e:
        std::replace(intypes.begin(), intypes.end(), 'd', 'e');
        if (tf = find_exact_tyfu(intypes))
            return tf;
        // cast all e to d:
        std::replace(intypes.begin(), intypes.end(), 'e', 'd');
        if (tf = find_exact_tyfu(intypes))
            return tf;
    }
    return NULL;
}


//**************************************************************************************************
//*  Class CFunc
//**************************************************************************************************

//! Return default coordinate of function of one argument.

CCoord CFunc::coord(const CCoord& co1) const
{
    CCoord co = co1;
    if (tag == "0-") {
        co.name = "-" + co1.name;
    } else if (tag == "ln") {
        co.name = "ln " + co1.name;
    } else if (tag == "lg") {
        co.name = "lg " + co1.name;
    } else if (tag == "sqrt") {
        co.name = "sqrt (" + co1.name + ")";
        co.unit = co1.unit == "" ? "" : "(" + co1.unit + ")^1/2";
    } else {
        co.name = tag + "(" + co1.name + ")";
    }
    return co;
}

//! Return default coordinate of function of two arguments.

CCoord CFunc::coord(const CCoord& co1, const CCoord& co2) const
{
    CCoord co = co1;

    if (tag == "+" || tag == "-") {
        if (co1.name == co2.name) {
            if (co1.unit != co2.unit)
                throw S("same coordinate name and different units in additive operation");
            co = co1;
        } else {
            co.name = co1.name + tag + co2.name;
            if (co1.unit == "")
                co.unit = "";
            else if (co2.unit != "" && co2.unit != co1.unit) {
                cout << "? different units " << co1.unit << " and " << co2.unit
                     << " in additive operation\n";
                co.unit = "";
            } else
                co.unit = co1.unit;
        }
    } else if (tag == "*" || tag == "/") {
        co.name = co1.name + tag + co2.name;
        if (co1.unit != "" && co2.unit != "")
            co.unit = co1.unit + tag + co2.unit;
        else if (co1.unit != "")
            co.unit = co1.unit;
        else if (co2.unit != "")
            co.unit = tag == "*" ? co2.unit : "1/(" + co2.unit + ")";
        else
            co.unit = "";
    } else if (tag == "^") {
        if (co2.unit != "") {
            cout << "? exponent is not dimensionless\n";
        }
        co.name = co1.name + "^" + co2.name;
        co.unit = co1.unit == "" ? "" : (co1.unit + "^<val>");
    } else {
        co.name = tag + "(" + co1.name + "," + co2.name + ")";
        co.unit = "";
    }
    return co;
}

//! Return default coordinate of function of three arguments.

CCoord CFunc::coord(const CCoord& co1, const CCoord& co2, const CCoord& co3) const
{
    CCoord co;
    co.name = tag + "(" + co1.name + "," + co2.name + "," + co3.name + ")";
    co.unit = "";
    return co;
}
