//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  registry.hpp
//! \brief declares a generic registry mixin interface.

#ifndef IREGISTRY_H
#define IREGISTRY_H

#include <map>
#include "../trivia/singleton.hpp"

//! Mixin interface for registries holding objects of type T; temporary version for SFuncRegistry.

template <class T> class IRegistryTmp
{
protected:
    virtual string type_name() const = 0;
    std::map<string, /*const*/ T*> FMap; //! unsorted hash, for expression evaluation
    vector<const T*> FList; //! sorted array, for help text
public:
    virtual ~IRegistryTmp()
    {
        for (const T* F : FList)
            delete F;
    }
    void push_back(const string& key, const T* val)
    {
        if (FMap.find(key) != FMap.end())
            throw "Duplicate registry entry " + key;
        FMap.insert(std::make_pair(key, val));
        FList.push_back(val);
    }
    const T* find(const string& key) const
    {
        auto pos = FMap.find(key);
        return pos == FMap.end() ? nullptr : pos->second;
    }
    const T* find_or_fail(const string& key) const
    {
        const T* ret = find(key);
        if (!ret)
            throw "Cannot find '" + key + "' in " + type_name() + " registry";
        return ret;
    }
};

//! Mixin interface for registries holding objects of type T

template <class T> class IRegistry
{
protected:
    virtual string type_name() const = 0;
    std::map<string, const T*> FMap; //! unsorted hash, for expression evaluation
    vector<const T*> FList; //! sorted array, for help text
public:
    virtual ~IRegistry()
    {
        for (const T* F : FList)
            delete F;
    }
    void push_back(const string& key, const T* val)
    {
        if (FMap.find(key) != FMap.end())
            throw "Duplicate registry entry " + key;
        FMap.insert(std::make_pair(key, val));
        FList.push_back(val);
    }
    const T* find(const string& key) const
    {
        auto pos = FMap.find(key);
        return pos == FMap.end() ? nullptr : pos->second;
    }
    const T* find_or_fail(const string& key) const
    {
        const T* ret = find(key);
        if (!ret)
            throw "Cannot find '" + key + "' in " + type_name() + " registry";
        return ret;
    }
};

#endif // IREGISTRY_H
